# =====================================================================
# csaCgi.awk: decode CGI variables.
#
# Copyright (c) 2002,2006 Carlo Strozzi
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 dated June, 1991.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# =====================================================================
# Notes:
#
# Decodes HTTP request parameters and turns them into valid rc(1)
# assignments. This script contains an AWK re-implementation and
# extension of the "uncgi" utility. See uncgi home page for more info:
# http://www.midwinter.com/~koreth/uncgi.html With respect to the
# original program, this script adds several CSA- and rc(1)- specific
# enhancements, and supports file-uploads.
# =====================================================================
# $Id: csaCgi.awk,v 1.2 2006/03/10 21:36:17 carlo Exp $

BEGIN {

   rc = 0

   # Safety checks.
   if (ENVIRON["TMPDIR"] !~ /^\// ||
       ENVIRON["libstem"] !~ /^\// ||
       _PID !~ /^[1-9][0-9]*$/) exit(rc=1)

   # Get file-names from env.
   o_file = ENVIRON["CSA_RPC_CMD"]
   awk_file = ENVIRON["CSA_RPC_AWK"]
   arg_file = ENVIRON["CSA_RDF_ARG"]
   put_file = ENVIRON["CSA_RPC_PUT"]

   first = 1

   lang = "en_US"

   print "BEGIN{\n_creat(ENVIRON[\"CSA_RPC_WWW\"],_O_TRUNC)" > awk_file

   # Supported national languages.

   delete _LANG

   _LANG["en"] = "en"
   _LANG["en_UK"] = "en_UK"
   _LANG["en_US"] = "en_US"
   _LANG["es"] = "es"
   _LANG["it"] = "it"

   # Misc. settings.
   rpcfile = map_file = "/dev/null"
   file_stem = ENVIRON["TMPDIR"] "/hexBinary." _PID

   # Max. input size, in bytes.
   rpc_maxsize = ENVIRON["CSA_RPC_MAXSIZE"]
   if (!rpc_maxsize) rpc_maxsize = 10000

   # First: decode HTTP cookie, if any.
   #
   # Note: when a cookie is set more than once in the client browser,
   # the latter returns the latest setting first. It will be up to the
   # calling rc(1) program to account for this and explicitly reference
   # the relevant rc(1) list element if appropriate.
   # Since cookie-based variables are taken into account first, their
   # values will be listed first in the resulting array, before
   # any form and URI values.
   # The order is: cookie, headers, POST, QUERY_STRING, PATH_INFO.

   if (ENVIRON["HTTP_COOKIE"] != "") {
      if (_uncgi(_hdrtocgi(ENVIRON["HTTP_COOKIE"]),o_file,1) == _EMSGSIZE) {

	 print "HTTP_COOKIE=()" > o_file		# shrink env.

	 # Exit with an error if the max. request message
	 # size is exceeded.

	 print "csaExit.fault 413" > o_file
      }
   }

   # Second: decode HTTP custom headers.
   #
   # HTTP custom headers are treated exactly the same as HTTP cookies.
   # In case of multiple settings for the same variable through
   # different means, the following order applies in the resulting rc(1)
   # list: cookie, headers, POST, QUERY_STRING, PATH_INFO. Special
   # characters passed in header values must be URI-encoded, as usual.

   for (env in ENVIRON) {
       e = env					# Save before mangling.
       if (!sub(/^HTTP_X_/,"",env)) continue
       if (env == "") continue			# re-check after mangling.

       if (_uncgi(_hdrtocgi("X-" env "=" ENVIRON[e]),o_file) == _EMSGSIZE) {

	  # Exit with an error if the max. request message
	  # size is exceeded.

	  print "csaExit.fault 413" > o_file
       }
       printf("HTTP_X_%s=()\n", env) > o_file	# shrink env.
   }

   if (ENVIRON["CONTENT_LENGTH"] > rpc_maxsize) {

      # Exit with an error if the max. input size is exceeded. This
      # is only a first check, that relies on the Content-Length HTTP
      # header. We do not control such header, although it can usually
      # be relied upon, but as a last-resort measure I check the actual
      # input size again further down, just in case.

      print "csaExit.fault 413" > o_file
   }

   # Skip confirmation in case of very large request body, to avoid
   # sending large objects back an forth between the server and the
   # client.

   else if (ENVIRON["CONTENT_LENGTH"] > 200000)
				print "CSA_CONFIRMED=1" > o_file
}

# DELETE must not take any content!
ENVIRON["REQUEST_METHOD"] == "DELETE" { next }

# Store the input entity body if "PUT". It can be either an HTTP or a
# CSA PUT (see the CSA front.end). In any case the entity body gets
# stored to a temporary file unchanged. Note that for a possible ending
# newline in the input not to be dropped, it is required that 'csa-escape -z'
# adds an extra trailing newline in the data itself. Such extra character
# will not affect the multipart format, due to the presence of the MIME
# part boundary in the relevant input.

ENVIRON["REQUEST_METHOD"] == "PUT" {

   if (!first) {
      if (hold) printf("\r\n") | _CMD
      else printf("\n") | _CMD
   }
   else _CMD = "exec csa-unescape -z > " put_file

   if (sub(/\r$/,"")) hold = 1
   else hold = 0
   printf("%s", $0) | _CMD
   first = 0
   next
}

# handle POST

ENVIRON["REQUEST_METHOD"] == "POST" {

   if (ENVIRON["CONTENT_TYPE"] == "application/x-www-form-urlencoded") {

      # MSIE 5.01 appends a trailing CR+LF to each input row of the POST
      # request body, and such trailer must be removed since it is not
      # to be considered as part of the input data. Note that since the
      # CSA front-end script now calls 'csa-escape' in any case, to fix
      # the bogus content-length field supplied by MSIE (again!), the
      # fix below may now be unnecessary, but it won't hurt either.

      if (ENVIRON["HTTP_USER_AGENT"] ~ / MSIE /) sub(/\r$/,"")

      if (input != "") input = input "\n"
      input = input $0
      next
   }

   # otherwise assume content-type is multipart/form-data

   if (NR == 1) {

      boundary = $0

      # Safety check.
      if ($0 !~ /^--.+\r$/) exit(rc=1)	# Not a MIME boundary.

      eof = $0
      sub(/\r$/,"",eof)
      eof = eof "--"
   }
}

# Note: the actual boundary is whatever string is indicated in the
# "boundary" attribute, prepended by two hyphens "--". That is, for
# "boundary=AaB03x" the actual boundary will be "--AaB03x", and the
# ending boundary will be "--AaB03x--".

# Note that to be rigorous, I should un-escape the backslashes that
# were introduced by 'csa-escape -z' also in MIME headers, but such
# characters are not normally present in those parts of the input
# message, so I can leave without unescaping them here.

$0 == boundary {name = ""; header = $0 "\n"; binary = 0; next}

$0 == eof || $0 == eof "\r" {exit}

header != "" {
   header = header $0 "\n"
   if ($0 ~ /^\r$/) {
      binary = 0
      name = header

      if (sub(/filename="[^"]+"/,"",header)) {
	 binary = 1
	 first = 1
      }

      header = name				# restore from backup
      name = ""

      #printf("%s", header)                  # debug

      # Use either the 'name=' or the 'filename=' values, whichever is
      # available. Usually it is 'name=', but as a last resort I will
      # use the value of 'filename=', and sanitize it further down.

      if (sub(/^.*: form-data; *name="/,"",header) || \
		sub(/^.*: attachment; *filename="/,"",header)) {

         sub(/".*/,"",header)
         name = header; norm = rec = 0

	 # Unlike uncgi, for names to be normalized
	 # they must begin with a dot.

         if (!binary) norm = sub(/^\./, "", name)

         # replace invalid characters with underscores (_)
	 # in name/filename value.

         gsub(/[^A-Za-z0-9_]/, "_", name)

         if (name == "") exit(rc=1)	# Unexpected input format.
         #print name                         # debug
      }
      header = ""
   }
   next
}

binary {
  if (!first) {
     if (hold) printf("\r\n") | _CMD
     else printf("\n") | _CMD
  }
  else {
     if (o_tmp != "") close(_CMD)
     o_tmp = file_stem "." name ".tmp"
     _cgivar(name,o_tmp,1)
     _CMD = "exec csa-unescape -z > " o_tmp 
  }
  if (sub(/\r$/,"")) hold = 1
  else hold = 0
  printf("%s", $0) | _CMD
  first = 0
  next
}

header == "" {
  #print			# debug
  sub(/\r$/,"")			# Remove trailing CR's
  if (norm) {
     gsub(/\r/,"\n")		# Turn embedded CR's into NL's.
     sub(/^[ \t]+/,"")		# Strip leading white-space.
  }

  rec++				# detect multiline input (textarea)

  # Un-escape backslashes that were escaped by "csa-escape -z" (see the CSA
  # main wrapper). Binary zeroes are not unescaped here, of course, or
  # mawk(1) would break, but on the other hand they are not expected in
  # STDIN input parts other than files.

  gsub(/\134\134/, "\134")

  # append new list element only if 1st (or only) record of input,
  # otherwise append to the same element.

  if (rec == 1) _cgivar(name,$0)
  else {
     _CGIVAL[name] = _CGIVAL[name] "\n" $0
     # multi-line values can only map to scalars.
     if (rec == 2) sub(/.?$/,"0",_RDFMAP[name])
  }
}

END {

   if (_CMD != "") close(_CMD)	# Close the PUT stream if open.

   # Decode POST variables, then GET variables, and finally PATH_INFO
   # variables last. Like uncgi, treat HTTP GET and POST the same.
   # Given the decoding order, GET overrides POST, and PATH_INFO
   # overrides GET. Note: due to this process, if a generic variable is
   # assigned multiple values in the three methods, those values will
   # be listed in reverse order in the resulting rc(1) list.
   # It is recommended that a variable be assigned with only one
   # mechanism.

   # Note: if urlencoded input comes from an XML attribute value, either
   # as PATH_INFO or QUERY_STRING data, it may be XML-encoded, so I
   # need to account for that. In fact, at lead in case of a Web browser,
   # the decoding *may* by done by the browser itself, but double-
   # checking won't hurt.

   ENVIRON["QUERY_STRING"] = _xmldecode(ENVIRON["QUERY_STRING"])
   ENVIRON["PATH_INFO"] = _xmldecode(ENVIRON["PATH_INFO"])

   # Apply the site-wide remapping rules. Another good location for
   # this file would be /etc.

   rpcappl = _pathinfo(ENVIRON["CSA_INSTALL"] "/etc/rest.map")

   # Apply the application-specific remapping rules. Note that if the
   # target application name cannot be worked out from the site-wide
   # mapping rules it must be present verbatim at the beginning of the
   # original PATH_INFO, so that the previous call to _pathinfo() gets
   # it anyway. If neither are the cases, then no remapping can be done
   # at all. The application-specific rewriting rules must not comprise
   # the application name again, or they won't match.

   if (rpcappl != "") {
      class_re = "^/" rpcappl "/+"
      sub(class_re,"/",ENVIRON["PATH_INFO"])
      _pathinfo(ENVIRON["libstem"] "/" rpcappl "/lib/rest.map")
      ENVIRON["PATH_INFO"] = "/" rpcappl ENVIRON["PATH_INFO"]
   }

   gsub(";", "/", ENVIRON["PATH_INFO"])		# see RFC-2396

   n = split(ENVIRON["PATH_INFO"], a, "/")

   if (ENVIRON["QUERY_STRING"] != "")
		ENVIRON["QUERY_STRING"] = ENVIRON["QUERY_STRING"] "&"

   # Consider class and method here only if present.
   if (n >= 3)
      ENVIRON["QUERY_STRING"] = ENVIRON["QUERY_STRING"] "0=" a[2] "." a[3] "&"

   for (i=4; i<=n; i++) {

     # Individual slash-delimited PATH_INFO tokens are treated like
     # HTTP custom headers and cookies if they contain the special
     # characters "=" or "&". Otherwise they are treated as
     # pseudo-positional arguments. Names that begin with "X-"
     # (case-insensitive) are turned into upper-case.

     if (a[i] ~ /[=&]/)
	ENVIRON["QUERY_STRING"] = ENVIRON["QUERY_STRING"] _hdrtocgi(a[i]) "&"

     else ENVIRON["QUERY_STRING"] = ENVIRON["QUERY_STRING"] i-3 "=" a[i] "&"
   }

   if (_uncgi(input,o_file) == _EMSGSIZE || \
	_uncgi(ENVIRON["QUERY_STRING"],o_file) == _EMSGSIZE) {

      # Exit with an error if the max. request message
      # size is exceeded.

      print "csaExit.fault 413" > o_file
   }

   # decode RwA payload first, if present.

   for (i in _CGIVAL)
       if (i == "CSA_RWA") _rwa2csa(_CGIVAL[i])

   # print the resulting assignments

   for (i in _CGIVAL) {

       if (i == "0" || i == "RwA" || i == "CSA_RWA") continue

       if (i == "X_CSA_CONFIRMED") {
	  _CGIVAL[i] = tolower(_CGIVAL[i])
	  if (_CGIVAL[i] ~ /^(y(es)?|1|true)$/)
		printf("CSA_CONFIRMED=1\n") > o_file

	  continue		# do not include in 'rwadiv'.
       }

       # Split multi-valued args and print RDF call statements. If an
       # associated RwA object exists, it means that it was detected
       # during a multipart POST operation (see further down). In this
       # case the parameter is expected to contain a single value in the
       # form of an absolute path to a temporary file where the binary
       # content was stored. The resulting RDF statement is thus set to
       # point at the relevant RwA attachment. Note that the resource
       # shown is a fake one, not to expose actual path names on the
       # server.

       printf("<param><rdf:Seq rdf:about=\"#call.%s\">\n",_xmlencode(i)) > arg_file

       n=split(_CGIVAL[i], a, "\001"); split(_RDFMAP[i],b,"\001")

       for (j=1; j<=n; j++) {
	 printf("<rdf:li rdf:") > arg_file
	 if (b[j]) {
	    printf("resource=\"") > arg_file

	    #if (b[j] == 1) printf("file:%s\" />\n",i) > arg_file
	    #else printf("%s\" />\n",a[j]) > arg_file

	    if (b[j] == 1) {
	       # Due to attachments and b64 auto-naming, the actual
	       # variable name can only be inferred from the associated
	       # temporary file name.
	       # To prevent large inline blocks of b64 data in the <call>
	       # section of the response, inlines are represented as pseudo
	       # RwA attachments.
	       sub(/.tmp$/,"",a[j]); sub(/.*\./,"",a[j])
	       printf("file:%s\" />\n",a[j]) > arg_file
	    }
	    else printf("%s\" />\n",a[j]) > arg_file
	 }

	 else {

	    # By default, use RDF (auto-detected) data types.
	    # Some applications may need to switch it off, to better
	    # reflect the nature of the CGI interface, where everything
	    # is just a string.

	    if (ENVIRON["CSA_RDF_TYP"] != 0)
	       printf("datatype=\"%s\">%s</rdf:li>\n",\
			_rdftype(a[j]),_xmlencode(a[j])) > arg_file
	    else printf("datatype=\"%s\">%s</rdf:li>\n",\
			_rdftype(_NULL),_xmlencode(a[j])) > arg_file
	 }
       }
       print "</rdf:Seq></param>" > arg_file

       # Handling method-calls may require data from other fields first,
       # so it needs to be done last. The drawback is that we will not
       # be able to determine the called application/method right
       # from the beginning, nor to assess in what user-level library,
       # if any, the application-level _userproc() function is stored.
       # That's why we need to store the WWW_xxx assignments in the
       # temporary AWK script 'awk_file', to be able to use them when
       # we know what method to call.

       if (i == "X_CSA_LANG") {
	  if (_LANG[_CGIVAL[i]] != "") {
	     lang = _CGIVAL[i]
	     printf("CSA_LANG=%s\n",_CGIVAL[i]) > o_file
	  }
       }

       else if (i == "X_CSA_RID") {
	  if (_CGIVAL[i] ~ /^[a-zA-Z]+_[1-9][0-9]*$/)
	     printf("CSA_RID=%s\n",_CGIVAL[i]) > o_file
       }

       else if (i == "X_CSA_GUID") {

	  # I accept a GUID value in any of the following formats:
	  #
	  # Name		Example
	  # ----		-------
	  # urn:oid		urn:oid:1.2.3...
	  # tagURI		tag:hawke.org,2001-06-05:Taiko
	  # RFC922 message-id	123456.123@example.org
	  # HTTP URL		http://www.example.org/123456/123
	  # GUID/UUID		30FED2A9-3C99-139B-04E7-1BE4290302B9
	  # GUID/UUID		8ddc491a-0fbf-11da-b685-004063d3b794
	  #
	  # Some of the above formats are more compact than others and
	  # I think they are therefore preferable. Currently I am
	  # biased in favour of the tagURI format. See http://taguri.org
	  # and http://taguri.org/draft-kindberg-tag-uri-04.txt

	  if (_CGIVAL[i] ~ /^[-_:,\/.a-zA-Z0-9]+(@[-.a-zA-Z0-9]+)?$/)
	     printf("CSA_GUID=%s\n",_CGIVAL[i]) > o_file
       }

       else if (i == "X_CSA_FMT") _rcset("CSA_RPC_FMT", _CGIVAL[i], o_file)

       else if (i ~ /^X_CSA_SESSION_[1-9]$/)
		_rcset(substr(i,3), substr(_CGIVAL[i],1,128), o_file)

       else if (i == "CSA_ISINDEX") _rcset(i, _CGIVAL[i], o_file)

       # The RWA message may be big, do not fully pass it to rc(1), not
       # to bloat the shell environment. Only the first few chars are
       # assigned, for documentational purposes.
       #
       #else if (i == "CSA_RWA")
       #	_rcset(i, substr(_CGIVAL[i],1,30) " ...",o_file)

       else if (_CGIVAL[i] != "") {

	  # Store variable for later processing.
	  printf("ENVIRON[\"WWW_%s\"]=%s\n",\
			i,_escape(_CGIVAL[i])) > awk_file
       }
   }

   # Get HTTP basic authorization data, if available.
   #
   # This header is set by the client browser in reply to a
   # "WWW-Authenticate" request from the server. It is decoded last so
   # that it will override authentications provided through other means,
   # like an HTTP cookie or an application-level token.

   _authset(ENVIRON["HTTP_AUTHORIZATION"],0,o_file)

   # Call the application-level processor and close the BEGIN{} block.
   print "_userproc(_O_REQUEST)}" > awk_file

   # Now handle the CSA-RPC method call, if any.
   # Truncate the method name to 64 characters and extract
   # application and program name.

   if (_CGIVAL["0"] != "") {

      rpc = substr(_CGIVAL["0"],1,64)
      rpcappl = rpc; sub(/\..*$/,"",rpcappl)
      rpcname = rpc; sub(/^[^.]+\./,"",rpcname)

      if (rpcappl ~ /^[a-zA-Z0-9][-_:a-zA-Z0-9]*$/ && \
          rpcname ~ /^[a-zA-Z0-9][-_.:a-zA-Z0-9]*$/ && \
	  rpcname !~ /\.(BEGIN|END)$/) {

	  csarc = ENVIRON["libstem"] "/" \
			rpcappl "/csa-" ENVIRON["CSA_VERSION"] ".rc"

          printf("CSA=%s\n",csarc) > o_file

	  # Switch to default profile if versioned one not found.

	  if (_stat(csarc)) {
	     csarc = ENVIRON["libstem"] "/" rpcappl "/csa.rc"
             printf("CSA=%s\n",csarc) > o_file
	  }

	  # This test is necessary, otherwise if also the default file is
	  # missing the error msg will be embedded into the HTTP response
	  # message.

	  if (_stat(csarc)) print "csaExit.fault 404" > o_file

	  printf("CSA_PGM=($CSA_PGM '%s')\n",rpcname) > o_file

          print ". $CSA || csaExit.fault" > o_file

	  # From now on, csaExit.fault() is the one provided by csaCgilib.rc,
	  # which was included by csaMainlib.rc.

	  # This can be useful for generating MIME multipart responses.
	  printf("CSA_MIME_BOUNDARY=%s\n", _mimebound()) > o_file

          print "csaUnlist CSA_WORKFILES $CSA_RPC_CMD $CSA_RPC_AWK $CSA_RDF_ARG $CSA_RPC_PUT " > o_file
          print "csaTrapFile $CSA_RPC_CMD $CSA_RPC_AWK $CSA_RDF_ARG $CSA_RPC_PUT " > o_file

          print "~ $CSA_DEBUG 0 || cat $CSA_RPC_CMD >[1=2]" > o_file

	  print "csaMkTemp CSA_RPC_WWW CSA_RDF_^(TOC BODY MRW)" > o_file

	  # Does the method-specific RDF file exist ? Note that unless the
	  # language code is specified as one of the CGI call arguments, the
	  # default "en_US" path will always be used here.

	  if (!_stat(ENVIRON["libstem"] "/" rpcappl "/doc/" lang "/" rpcname ".rdf")) {
	     print "CSA_RPC_IDL=" ENVIRON["libstem"] "/" rpcappl "/doc/" lang "/" rpcname ".rdf" > o_file
	  }

	  # Does the application-level argument-processing AWK
	  # library exist ? Otherwise use the CSA default library.
	  # Note that the user program has no way to specify
	  # additional libraries. If they are needed, they will
	  # have to be statically included in the user-level library.
	  # If the user-level library exists, it must define at least
	  # _userproc().

	  if (!_stat(ENVIRON["libstem"] "/" rpcappl "/lib/rpcio/" rpcname ".awk")) {
	     # Set application-specific I/O library.
	     print "CSA_RPC_IOLIB=" ENVIRON["libstem"] "/" rpcappl "/lib/rpcio/" rpcname ".awk" > o_file
	  }
	  else {
	     # Set default I/O library.
	     print "CSA_RPC_IOLIB=$CSA_INSTALL/lib/csa-rpciolib.awk" > o_file
	  }

	  # At this point CSA_RPC_WWW should have been written.
	  # Nope, don't run the generated code automatically, as
	  # it is up to CSA_PGM to decide when/if to load args.
	  #print ". $CSA_RPC_WWW || csaExit.fault 550 0008 $CSA_RPC_WWW" > o_file

          rpcfile = ENVIRON["libstem"] "/" rpcappl "/rpclib/" rpcname

	  # The method.BEGIN file, if any, must be called before 'csaAwk',
	  # so that any shell-level stuff that may be needed by _userproc()
	  # can be prepared by the programmer beforehand. Like AWK's BEGIN{}
	  # special function, the BEGIN script is run *before* any request
	  # input data is read.

	  if (_stat(rpcfile ".BEGIN") == _TRUE)
		print ". " rpcfile ".BEGIN" > o_file

	  print "csaIsFullPath --quiet --exist $CSA_ROOT/lib/i18n/$CSA_LANG.awk &&" > o_file
	  print "  CSA_RPC_IOLIB=($CSA_RPC_IOLIB $CSA_ROOT/lib/i18n/$CSA_LANG.awk)" > o_file 

	  # This will write rc(1) assignments to CSA_RPC_WWW.
	  printf("csaAwk %s -- -- /dev/null || csaExit.fault 550 0003 csaAwk\n",awk_file) > o_file
      }
   }

   if (_stat(rpcfile) == _TRUE) {
      print ". " rpcfile > o_file
      print "csaExit.ok # safety valve" > o_file # fail-safe default.
   }
   else {

      # Requested method not found.
      print "CSA_PGM=$CSA_PGM(1); csaPrintMsg 0027 " \
		_rcescape(ENVIRON["CSA_REQUEST_URI"]) > o_file
      print "csaExit.fault 404 0067" > o_file
   }
}

# =====================================================================
# End of program
# =====================================================================
